/*  XR.C  a cross reference generator for C, Turbo Pascal, and Basic.
/*  Based on a 'C' and 'PASCAL' cross reference program from 6/86 Computer Language Magazine.
    Extensively redone by Bob Montgomery 9-8-88
					 
Language: Microsoft C ver 3.0 up.
Note:   Uses recursive functions, so when linking, you must include the LINK switch '/STACK:10000'.

The program uses a 'symrec' structure for each symbol found, a 'numrec' structure for each line number found, and
an array of null terminated strings for the symbol names. There is also an array of pointers to built in symbols
for the language.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <io.h>
#include <malloc.h>

#define TRUE       1
#define FALSE      0
#define MAX_DEF    500                  /* maximum language definitions */
#define NAME_CHARS  "._"                /* punctuation allowed in Names */
#define BAS_CHAR    "%&!#$"             /* BASIC variable size identifiers */
#define DQUOTE     0x22                 /* ASCII Double Quote Character */
#define SQUOTE     0x27                 /* ASCII Single Quote Character */
FILE *infile = NULL, *outfile = NULL;   /* Input & output files */
char linebuf[50];                       /* current symbol holder */
char inname[50];                        /* Input file name */
int  linenum     = 1;                   /* line counter */
int  lineptr     = 0;                   /* Assume empty line buffer */
int  is_c = 0, is_bas = 0, is_pas = 0;  /* Language flags */
int  printline   = 0;                   /* Current line on page */
int  num_per_line;                      /* Line numbers to print/line */
int  dupflag = 0, functflag = 0, varflag = 0;
int  linewidth = 132, pagelength = 80, leftmargin = 5;

struct linenumber_record                /* Linked list of line numbers for a symbol */
{   int number;                         /* Line number of a symbol occurance */
    struct linenumber_record *next;     /* Pointer to next line number record for same symbol */
};

struct symbol_record                    /* Binary tree record for a symbol */
{   char *symbol;                       /* Symbol name string pointer */
    int  symcount;                      /* Symbol occurance counter */
    struct linenumber_record *lnum;     /* Pointer to 1st line number record for this symbol */
    struct symbol_record *lnode;        /* Pointer to left branch of binary tree */
    struct symbol_record *rnode;        /* Pointer to right branch of binary tree */
};

struct symbol_record *symtab = NULL;    /* Declare symbol record pointer */
int symbols = 0;                        /* Number of symbols found */
char *Exceptions[ MAX_DEF ];            /* Language keyword table */
struct symbol_record *salloc();         /* Allocate symbol record memory function */
struct linenumber_record *nalloc();     /* Allocate line number record memory */
char *strsave();                        /* Allocate memory and save a string there */

main(argc, argv)
char *argv[];
int argc;
{    int  ch, i;
     char *ptr;

     for( i=1; i < argc; ++i )          /* Parse the command line */
     {    ptr = argv[i];                /* ptr --> i'th arg token */
          if( *ptr++ == '/' )           /* if it's a switch token */
          {    while (ch = *ptr++)      /* Get next char */
               {    switch(toupper(ch))      /* Process 'em */
                    {    case 'O':           /* Output file */
                              if (++i >= argc) help();  /* File name is next arg */
                              if ((outfile = fopen(argv[i], "w")) == NULL)
                              {    printf("ERROR: Cannot open output file - %s\n",argv[i]);
                                   exit(-2);
                              }
                              goto A1;  /* Next argv */
                         case 'W':           /* Chars/line for output (default 132) */
                              if (sscanf(ptr, "%d", &linewidth) != 1 || linewidth < 1 || linewidth > 200)
                              {    printf("Bad output line width (1 to 200 allowed)\n");
                                   help();
                              }
                              goto A1;  /* Next argv */
                         case 'L':           /* Lines/page for output (default 80) */
                              if (sscanf(ptr, "%d", &pagelength) != 1 || pagelength < 1 || pagelength > 200)
                              {    printf("Bad output pqge length (1 to 200 allowed)\n");
                                   help();
                              }
                              goto A1;  /* Next argv */
                         case 'M':           /* Left margin for output (default 5) */
                              if (sscanf(ptr, "%d", &leftmargin) != 1 || leftmargin < 0 || leftmargin > (linewidth - 1))
                              {    printf("Bad left margin (1 to linewidth-1 allowed)\n");
                                   help();
                              }
                              goto A1;  /* Next argv */
                         case 'D':           /* Duplicate line numbers switch */
                              dupflag++;
                              break;
                         case 'V':           /* Variables only switch */
                              varflag++;
                              break;
                         case 'F':           /* Functions only switch */
                              functflag++;
                              break;
                         default :           /* Bad switch, give help */
                              help();
                    }
               }
          }
          else
          {    if ((infile = fopen(argv[i], "r")) == NULL)
               {    printf("ERROR: Cannot open input file - %s\n",argv[i]);
                    exit(-2);
               }
               else strcpy(inname, argv[i]);
          }
A1: ;
     }                                  /* Done with cmd string */

    if (infile == NULL) help();         /* Need an input file */
    if (outfile == NULL) outfile = stdout; /* Default output is stdout */
    num_per_line = (linewidth - leftmargin) / 5;
    load_tbl();                         /* Load the table of keywords */
    printf("Getting symbols for %s\n", inname);
    while ( (ch=getc(infile)) != EOF ) filter(ch);  /* Get all symbols */
    fprintf( outfile, "     Symbol Table for File - %s\n", inname);
    printline++;                        /* Inc line number */
    printf("Writing symbols for %s\n", inname);
    symprint(symtab);                   /* Print all symbols */
    puts("Done");                       /* Print end, and quit */
    exit(0);                            /* Close files, free memory */
}
/* ========================================== */
filter(ch)     /* Remove comments, quoted text, and punctuation */
char ch;
{    char cl, temp[4];                  /* Test chars */
     static is_com = FALSE;             /* Start with comment flag false */

     if( ch==DQUOTE || (ch == SQUOTE && !is_bas))  /* Remove all between " or ' */
     {    while ( (cl=getc(infile)) != ch)
          {    if(cl == EOF) return;
               if (is_bas && cl == '\n')          /* Can get sloppy in BASIC */
               {    linenum++;
                    break;
               }
          }
          next_char( ' ' );             /* Cancel word in linebuf */
     }

     if (is_bas)                        /* If BASIC, check for a comment */
     {    if (ch == SQUOTE) is_com = TRUE;         /* ' is start of a comment */
          else if (ch == ' ')                      /* Check for REM */
          {    if (lineptr == 3)                       /* If 3 chars */
               {    linebuf[lineptr] = '\0';                /* Terminate string */
                    strcpy(temp, linebuf);                  /* Copy to temp */
                    if (!strcmp(strupr(temp), "REM")) is_com = TRUE;
                    lineptr = 0;                            /* Don't save the REM */
               }
          }
          else if (strchr(BAS_CHAR, ch) != NULL)  /* If %&!#$ variable size chars */
               if (lineptr) next_char(ch);             /* Tack on if variable in linebuf */
          while (is_com)                          /* If a comment */
          {    if ((ch = getc(infile)) == EOF) return;
               if (ch == '\n') is_com = FALSE;         /* Get chars to end of line */
          }
     }

     else if (is_pas)                             /* If PASCAL, check for comment */
     {    if((ch=='{'))                                /* If {, find } */
          {    while ((ch=getc(infile)) != '}' )
               {    if(ch == EOF) return;
                    if (ch == '\n') linenum++;
               }
               next_char(' ');
          }
          else if (ch == '(')                     /* Check for Pascal function or comment start */
          {    if (lineptr)                       /* If something in linebuf */
               {    linebuf[lineptr] = '\0';           /* Terminate whats there */
                    if (is_lang(linebuf))              /* If PASCAL built in symbol */
                    {    lineptr = 0;                       /* Reset linebuf pointer */
                         return;                            /* and ignore it */
                    }
                    else                               /* else add () to linebuf */
                    {    next_char('(');
                         next_char(')');
                         next_char(' ');                    /* and save symbol */
                    }
               }
               else
               {    if ((cl = getc(infile)) == EOF) return; /* Get next char */
                    if(cl != '*')                           /* If not *, put it back */
                    {    next_char(' ');
                         ungetc(cl, infile);
                    }
                    else                               /* If *, is start of comment */
                    {    is_com = TRUE;                     /* Set comment flag */
                         while (is_com)                     /* While in a comment */
                         {    while ((ch=getc(infile)) != '*' )
                              {    if(ch==EOF) return;
                                   if (ch == '\n') linenum++; /* If newline, inc line number */
                              }
                              if ((cl = getc(infile)) == EOF) return;
                              if(cl==')') is_com = FALSE;
                              else ungetc(cl, infile);
                         }
                         next_char(' ');
                    }
               }
          }
     }

     else if (is_c)
     {    if (ch=='/')                       /* Check for C comment start */
          {    cl=getc(infile);                   /* Get next char */
               if (cl==EOF) return;
               if(cl != '*')                      /* If not *, put it back */
               {    next_char(' ');               /* Check linebuf */
                    ungetc(cl, infile);           /* and put char back */
               }
               else                          /* else is start of comment */
               {    is_com = TRUE;
                    while (is_com)                /* While in a comment */
                    {    while ((ch=getc(infile)) != '*')  /* Look for * */
                         {    if (ch == EOF) return;
                              if (ch == '\n') linenum++;   /* If newline, inc line number */
                         }
                         if ((cl = getc(infile)) == EOF) return;  /* When * found, get next char */
                         if(cl=='/') is_com = FALSE;   /* If /, end of comment */
                         else ungetc(cl, infile);      /* else put it back */
                    }
                    next_char(' ');               /* Check linebuf */
               }
          }
          else if (ch == '(')                /* If a C function symbol */
          {    if (lineptr)                       /* If something in linebuf */
               {    linebuf[lineptr] = '\0';           /* Terminate whats there */
                    if (is_lang(linebuf))              /* If C built in symbol */
                    {    lineptr = 0;                       /* Reset linebuf pointer */
                         return;                            /* and ignore it */
                    }
                    else                               /* else add () to linebuf */
                    {    next_char('(');
                         next_char(')');
                         next_char(' ');                    /* and save symbol */
                    }
               }
          }
     }

     /* If A-Z, a-z, 0-9 or . or _ or space or cr add to linebuf */
     if (isalnum(ch) || strchr(NAME_CHARS, ch) || isspace(ch) ) next_char(ch);
     else if(ch=='-')                   /* Check for C -> symbol */
     {    if('>' == (ch=getc(infile)) )      /* and a > */
          {    next_char('-');               /* Print them to linebuf */
               next_char('>');
          }
          else                               /* else put the > back */
          {    ungetc(ch, infile);
               next_char(' ');
          }
     }
     else next_char(' '); /* Takes care of all other punctuation (/, *, +, -, (, ), %, etc) */
}
/* ======================================================== */
next_char(ch)    /* Add a char to line buffer; if space, reset pointer & update symbol table */
char ch;
{    if(ch == ' ' || ch == '\n')        /* If char is a space or newline */
     {    if(lineptr != 0)                   /* If buffer not empty */
          {    linebuf[lineptr] = '\0';           /* Add a null terminator */
               lineptr = 0;                       /* Reset line char counter */
               add_sym(linebuf, linenum);         /* Update or add 'name in linebuf' symbol record */
          }
          if(ch == '\n') ++linenum;          /* If char was newline, inc line number */
     }
     else linebuf[lineptr++] = ch;      /* else add char to line buffer */
}
/* ====================================================== */
add_sym(string, linenumber)   /* Add or update a symbol record */
char *string;
int linenumber;
{    struct symbol_record *oldsym, *newsym;
     struct linenumber_record *linerec;
     int  cond, foundflag = 0;
	
     if (is_pas || is_bas) strupr(string); /* Pascal and Basic not case sensitive */
     if (!isalpha(string[0])) return;   /* Return if 1st char is not a letter */
     if (is_lang(string)) return;       /* or if a built in language symbol */
     newsym = symtab;                   /* Start at root entry */
     while (newsym != NULL)             /* Find match, or end of tree */
     {    oldsym = newsym;                   /* Remember parent record */
          if ((cond = strcmp(string, oldsym->symbol)) == 0) /* If same name */
          {    foundflag++;                                 /* Set found flag */
               break;                                       /* and break out of loop */
          }
          else if (cond < 0) newsym = oldsym->lnode;   /* If less alphabetically go left */
          else newsym = oldsym->rnode;                 /* If greater alphabetically go right */
     }
     if (foundflag)                                    /* If symbol was found */
     {    oldsym->symcount++;                               /* Inc occurance count */
          linerec = oldsym->lnum;                           /* Get pointer to 1st line number structure */
          while (1)
          {    if (!dupflag && linerec->number == linenumber)  /* Return if duplicate line number */
                    return;
               if ((linerec->next) == NULL) break;          /* Quit if end of chain */
               else linerec = linerec->next;                /* else go to next line record */
          }
          linerec = linerec->next = nalloc();               /* Allocate memory for new line record */
     }
     else                                              /* If a new symbol */
     {    newsym = salloc();                                /* Allocate memory for new symbol structure */
          if (symbols)                                      /* If not 1st symbol */
          {    if (cond < 0) oldsym->lnode = newsym;             /* If less, is left node */
               else oldsym->rnode = newsym;                      /*  If greater, is right node */
          }
          else symtab = newsym;                             /* If 1st, is root symbol */
          newsym->symbol = strsave(string);                 /* Save string and point to it */
          newsym->symcount = 1;                             /* Count is 1 for this new symbol */
          newsym->lnode = newsym->rnode = NULL;             /* Right & left node pointers are null */
          newsym->lnum = linerec = nalloc();                /* Save line number record and point to it */
          symbols++;                                        /* Inc symbol count */
     }
     linerec->number = linenumber;                     /* Fill in new line record structure */
     linerec->next = NULL;
}
/* ====================================================== */
struct symbol_record *salloc()  /* Allocate memory for a symbol record */
{   struct symbol_record *recptr;

    if ((recptr = (struct symbol_record *)malloc(sizeof(struct symbol_record))) == NULL)
    {    puts("Out of symbol table space"); exit(-1); }
    return (recptr);
}
/* ======================================================== */
struct linenumber_record *nalloc()  /* Allocate memory for a line record */
{   struct linenumber_record *linptr;

    if ((linptr = (struct linenumber_record *)malloc(sizeof(struct linenumber_record))) == NULL)
    {    puts("Out of line number table space"); exit(-1); }
    return (linptr);
}
/* ======================================================= */
char *strsave(string)  /* Allocate memory for a string, and copy string to it */
char *string;
{    char *ptr;
     int size;

     size = strlen(string);             /* Get string length */
     if ((ptr = (char *)malloc(size + 1)) == NULL)
     {    printf("ERROR: Out of string space.\n");
          exit(-1);
     }
     else strcpy(ptr, string);          /* Copy string to new memory */
     return(ptr);                       /* Return pointer to new string */
}
/* ================================================== */
symprint(recptr)    /* This is a recursive symbol print routine */
struct symbol_record *recptr;
{    if (recptr != NULL)                /* If this is a record */
     {    symprint(recptr->lnode);                     /* Go left till a null found */
          if (functflag && strstr(recptr->symbol, "()") == NULL) goto B1;
          if (varflag && strstr(recptr->symbol, "()") != NULL) goto B1;
          margin();                                    /* When back, print a margin and symbol data */
          fprintf(outfile, "%-31s (%d)\n",recptr->symbol, recptr->symcount);
/* If symbol is end of page, go ahead and print line numbers if only 1 line long, else do form feed */
          if (++printline >= pagelength && howlong(recptr->lnum) > num_per_line) formfeed();
          numprint(recptr->lnum, 0);                   /* Print the line numbers */
          if (++printline >= pagelength) formfeed();   /* If end of line */
B1:       symprint(recptr->rnode);                     /* Then go right */
     }
}
/* =============================================== */
numprint(linptr, i)      /* This is a recursive line number print routine */
struct linenumber_record *linptr;
{    if(linptr != NULL)                 /* If this is a line record */
     {    if((i % num_per_line) == 0)        /* If 1st number or last on a line */
          {    if (i)                             /* If last number on this line */
               {    if (++printline >= pagelength) formfeed();  /* If end of page */
                    else putc('\n', outfile);          /* else send a new line if not 1st line number */
               }
               margin();                               /* Do a margin in both cases */
          }
          fprintf(outfile, " %4d", linptr->number);    /* Print the line number */
          numprint(linptr->next, ++i);                 /* Print the next line number */
     }
     else putc('\n', outfile);          /* Add a cr at end of numbers */
}
/* ================================================= */
is_lang(string)   /* See if a string is a built in language word; FALSE=>No, TRUE=>Yes */
char *string;
{   int i, j;
     char temp[50];

    i = FALSE;                               /* Assume not in table */
    strupr(strcpy(temp,string));             /* Copy to temp and make upper case */
    while (Exceptions[i] != NULL)            /* Check all built in symbols */
    {   if((j = strcmp(temp, Exceptions[i++])) == 0) return (TRUE); /* If found, return true */
	}
    return (FALSE);                          /* else return false */
}
/* ====================================================== */
load_tbl()          /* Load the language table to memory */
{    char *newptr, *oldptr, ch;
     char fname[65], symbol[50];
     FILE *tfile;
     int i;

     oldptr = strupr(inname);                     /* Point to input filename */
     while ((newptr = strchr(oldptr, '.')) != NULL) oldptr = newptr + 1;  /* Find last . */
     if ((i = strlen(oldptr)) > 3 || !i)          /* If no ext or ext > 3 chars */
     {    printf("ERROR: No extension on input file name: %s\n",inname);
          exit(-1);
     }
     if (!strcmp(oldptr, "C")) is_c++;            /* If ext is C */
     if (!strcmp(oldptr, "BAS")) is_bas++;        /* If ext is BAS */
     if (!strcmp(oldptr, "PAS")) is_pas++;        /* If ext is PAS */
     strcat(strcpy(fname, oldptr), ".XRF");       /* Copy ext to fname, and add ext .XRF */
     if ((tfile = fopen(fname, "r")) == NULL)     /* Open language built in symbol file */
     {   printf("ERROR: Cannot open Language Table file: %s\n",fname);
         exit(-2);
     }
     for (i=0; i < MAX_DEF; i++)                  /* Copy all names to memory */
     {    if(fscanf(tfile, "%s", symbol) == EOF)       /* If last keyword */
          {    fclose(tfile);                               /* Close the language file */
               printf("Language Table Loaded\n");           /* Print message */
               return;                                      /* and return */
          }
          else if (symbol[0] == '#')                   /* If a comment line */
          {    while (getc(tfile) != 10) ;                  /* Find end of line */
               i--;                                         /* Dec lang counter */
          }
          else Exceptions[i] = strsave(strupr(symbol)); /* Save to memory in upper case, add pointer to array */
     }
     printf("ERROR: Language Definition File is too long (Max = %d).\n",MAX_DEF);
     exit(-3);
}
/* ============================================ */
margin()            /* Print a left margin */
{    fputs("     ", outfile);           /* Print a margin */
}
/* ============================================ */
formfeed()          /* Do a formfeed */
{    putc(12, outfile);                 /* Send a form feed */
     printline = 0;                     /* Reset printline */
}
/* =========================================== */
howlong(lptr)       /* Find how many line numbers for this symbol */
struct linenumber_record *lptr;
{    int  i=1;

     while ((lptr = lptr->next) != NULL) i++;
     return (i);
}
/* ============================================ */
help()              /* Print help info */
{    printf("Usage is: XR [options] input_name.ext [/o output name]\n");
     printf("Options:  /wn sets listing width to n chars/line (132 default)\n");
     printf("          /ln sets listing length to n lines/page (80 default)\n");
     printf("          /mn sets left margin to n characters (5 default)\n");
     printf("          /dfv  d=print duplicate line #s, v=variables only, f=functions only\n");
     exit(0);
}
